#!/usr/sbin/perl -w
#
# nodeNumberer.pl v1.0
# Last modified March 26, 2004 15:30
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#	A nexus-formatted treefile
#
# Output:
#   A nexus-formatted treefile with node numbers contained as node labels (can be viewed, for
#	example in TreeView)
#
# Usage: nodeNumberer.pl [-t<filename>] [-p] [-v] [-h]
#   options: -t<filename> = treefile containing trees in NEXUS format
#                           (default = treefile.tre)
#            -p = make output suitable for PAUP*
#            -v = verbose output
#            -h = print this message and quit

use strict;

# Set defaults
	# Input variables
		my $treefile = "treefile.tre";
		my $block = 1;
		my $treeBlock = 0;
	
	# Tree variables
		my $SupportFlag;

	# Output variables
		my $paupOutput = 0;

	# Miscellaneous variables
		my ($startTime, $totalTime);
		my $verbose = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
		{
		$treefile = $1;
		}
	elsif ($ARGV[$i] =~ /-p/)
		{
		$paupOutput = 1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: nodeNumberer.pl [-t<filename>] [-p] [-v] [-h]\n";
		print "Options: -t<filename> = file containing trees in NEXUS format\n";
		print "                        (default = treefile.tre)\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: nodeNumberer.pl [-t<filename>] [-p] [-v] [-h]\n";
		exit(1); 
		}
	}

(my $outFile = $treefile) =~ s/\.\w*$//;
	$outFile .= "_numbered.tre";

print "Reading in tree descriptions ...\n";
	$startTime = time;
	setLineBreak($treefile);
	open (fTREES, "<$treefile") or die "Cannot open treefile $treefile\n";		# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
		my (@AllTrees, %treeNames, $treeDescrip, $treeFlag, $treeCount);
		my (@taxonList, %taxonPresent);
		my $NumChar = 0;
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<fTREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$treeDescrip .= $_;
				if ($_ =~ m/:/)
					{
					$SupportFlag = 1;
					}
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		#Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		#Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		#Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						push @AllTrees, $tree;
					print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close fTREES;

	$totalTime = time - $startTime;
	print "\tTime taken: $totalTime seconds\n";
	die "\nERROR: No trees read in from file $treefile\n" if (not @AllTrees);

# Print nexus-formatted tree file of supertree with labels of node numbers
	print "Outputting tree descriptions to $outFile ...\n";
	open (TREE, ">$outFile") or die "Cannot open output file $outFile to write tree description.\n";
		print TREE "#NEXUS\n\nbegin trees;\n";
		$treeCount = 0;
		foreach my $tree (@AllTrees)
			{
			$treeCount++;
			my ($position, $CurrentClade) = (0, 0);
			my (%ParentClade, %open);
			my $CladeID = 0;
			my $newST;
			my $OldCurrent;
			while ($position < length($tree))
				{
				my $offset = 1;
				if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
					{
					$CladeID++;
					$ParentClade{$CladeID} = $CurrentClade;
					$CurrentClade = $CladeID;
					$newST .= "(";
					}
				elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};
					$newST .= ")";
					my $label;
					if ($OldCurrent == 0)
						{
						$label = "root";
						}
					else
						{
						$label = "$OldCurrent'";
						if ($paupOutput)
							{
							$label = "'node".$label;
							}
						else
							{
							$label = "'".$label;
							}
						$newST .= $label;
						}
					}
				else
					{
					$newST .= substr($tree, $position, 1);
					}
				$position++;
				}
			print TREE "\tTREE $treeNames{$treeCount} = $newST;\n";
			}
		print TREE "end;\n";
	close TREE;

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}
	